/*
 * deviced
 *
 * Copyright (c) 2012 - 2013 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <stdbool.h>
#include <vconf.h>
#include <device-node.h>
#include <bundle.h>
#include <eventsystem.h>
#include <hw/battery.h>
#include <sys/stat.h>

#include "core/devices.h"
#include "core/device-notifier.h"
#include "core/udev.h"
#include "core/log.h"
#include "core/config-parser.h"
#include "display/poll.h"
#include "display/setting.h"
#include "shared/eventsystem.h"
#include "display/core.h"
#include "display/display-ops.h"
#include "apps/apps.h"
#include "power-supply.h"
#include "battery.h"
#include "extcon/extcon.h"

#ifndef VCONFKEY_SETAPPL_BLOCKMODE_WEARABLE_BOOL
#define VCONFKEY_SETAPPL_BLOCKMODE_WEARABLE_BOOL "db/setting/blockmode_wearable"
#endif

#ifndef VCONFKEY_SETAPPL_THEATER_MODE_ENABLE
#define VCONFKEY_SETAPPL_THEATER_MODE_ENABLE "db/setting/theater_mode_enable"
#endif

#ifndef VCONFKEY_SETAPPL_GOODNIGHT_MODE_ENABLE
#define VCONFKEY_SETAPPL_GOODNIGHT_MODE_ENABLE "db/setting/goodnight_mode_enable"
#endif

#define BATTERY_NAME        "battery"
#define CHARGEFULL_NAME     "Full"
#define CHARGENOW_NAME      "Charging"
#define DISCHARGE_NAME      "Discharging"
#define NOTCHARGE_NAME      "Not charging"
#define OVERHEAT_NAME       "Overheat"
#define TEMPCOLD_NAME       "Cold"
#define OVERVOLT_NAME       "Over voltage"
#define GOOD_NAME           "Good"

#define REMOVE_POPUP        "remove_battery_popups"
#define DISCONNECT_POPUP    "battdisconnect"

#define CHARGER_WIRELESS_TYPE_BT    10
#define CHARGER_WIRELESS_TYPE_3G    22
#define CHARGER_INCOMPATIBLE_TYPE   11
#define WIRELESS_CHARGER_CONNECTED  2

#define SIGNAL_CHARGEERR_RESPONSE "ChargeErrResponse"
#define SIGNAL_TEMP_GOOD          "TempGood"

#define ABNORMAL_CHECK_TIMER_INTERVAL 60

#define METHOD_FULL_NOTI_ON   "BatteryFullNotiOn"
#define METHOD_FULL_NOTI_OFF  "BatteryFullNotiOff"
#define METHOD_CHARGE_NOTI_ON "BatteryChargeNotiOn"

#define CHARGER_CHARGE_ON_NODE "/sys/class/power_supply/battery/chg_on"

#define RETRY_MAX 5
#define BATTERY_CHECK_TIMER_INTERVAL 500  /* 0.5 second */
#define LCD_DIM_TIME_IN_BATTERY_HEALTH	10000	/* ms */
#define HEALTH_POPUP_TERMINATE_TIMEOUT	10
#define HEALTH_POPUP_RELAUNCH_TIMEOUT	3600

enum charger_type {
	CHARGER_TYPE_NONE = 0,
	CHARGER_TYPE_WIRE,
	CHARGER_TYPE_WIRELESS,
	CHARGER_TYPE_INCOMPATIBLE,
};

static void uevent_power_handler(struct udev_device *dev);
static const struct uevent_handler uh = {
	.subsystem   = POWER_SUBSYSTEM,
	.uevent_func = uevent_power_handler,
};

struct battery_status battery;
struct battery_status old_battery;
static int noti_id;
static int online_status;

static guint power_timer;
static guint abnormal_timer;
static guint health_popup_timer;
static struct timespec health_relaunch_start_st;
static struct timespec health_relaunch_end_st;

bool battery_do_not_disturb(void);
static int battery_pm_change_internal(int pid, int s_bits);
static int display_changed(void *data);
static void health_popup_timer_init(const char *func);
static void update_health_popup_relaunch_time(int timeout);
static void change_health_popup_timer(enum battery_noti_status status);
static int booting_done(void *data);

static struct battery_device *battery_dev;

static int load_uevent(struct parse_result *result, void *user_data);
static int event_handler_state_changed(void *data);

static int lowbat_execute(void *data)
{
	static const struct device_ops *lowbat_ops;

	FIND_DEVICE_INT(lowbat_ops, "lowbat");
	return device_execute(lowbat_ops, data);
}

static int power_supply_broadcast_str(char *sig, char *status)
{
	static char old_sig[32];
	static char old_status[32];
	char *str;
	int ret;

	if (!sig || !status) {
		_E("there is no signal name");
		return -EINVAL;
	}

	if (strncmp(old_sig, sig, strlen(sig)) == 0 && strncmp(old_status, status, strlen(status)) == 0)
		return 0;

	snprintf(old_sig, sizeof(old_sig), "%s", sig);
	snprintf(old_status, sizeof(old_status), "%s", status);

	str = status;

	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_BATTERY,
						DEVICED_INTERFACE_BATTERY,
						sig,
						g_variant_new("(s)", str));

	if (ret == 0)
		return 1;
	else
		return ret;
}

static void pm_check_and_change(int bInserted)
{
	static int old = -1;

	if (old == bInserted)
		return;
	old = bInserted;
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
}

static void abnormal_popup_timer_init(void)
{
	if (abnormal_timer == 0)
		return;
	g_source_remove(abnormal_timer);
	abnormal_timer = 0;
	_I("Delete health timer.");
}

static void health_status_broadcast(void)
{
	dbus_handle_emit_dbus_signal(NULL,
		DEVICED_PATH_BATTERY,
		DEVICED_INTERFACE_BATTERY,
		SIGNAL_TEMP_GOOD,
		NULL);
}


static void health_timer_reset(void)
{
	abnormal_timer = 0;
}

static gboolean health_timer_cb(void *data)
{
	health_timer_reset();

	if (battery.health != HEALTH_LOW && battery.health != HEALTH_HIGH)
		return G_SOURCE_REMOVE;

	_I("Popup: Battery health status is not good.");
	device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
	if (disp_plgn.pm_unlock_internal)
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	if (battery.health == HEALTH_LOW)
		battery_charge_err_low_act(NULL);
	else if (battery.health == HEALTH_HIGH)
		battery_charge_err_high_act(NULL);
	return G_SOURCE_REMOVE;
}

static void abnormal_popup_dbus_signal_handler(GDBusConnection  *conn,
		const gchar      *sender,
		const gchar      *path,
		const gchar      *iface,
		const gchar      *name,
		GVariant         *param,
		gpointer          user_data)

{
	if (battery.health == HEALTH_GOOD)
		return;
	_I("Restart health timer.");
	abnormal_timer = g_timeout_add_seconds(ABNORMAL_CHECK_TIMER_INTERVAL,
						health_timer_cb, NULL);
	if (abnormal_timer == 0)
		_E("Failed to add abnormal check timer.");
}

static void full_noti_cb(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to notify full: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	noti_id = id;
	_D("Inserted battery full noti(%d).", noti_id);

out:
	g_variant_unref(var);
}

static void noti_off_cb(GVariant *var, void *user_data, GError *err)
{
	int ret = 0;

	if (!dh_get_param_from_var(var, "(i)", &ret)) {
		_E("Failed to off notification: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	_D("Noti off: %d", ret);

out:
	g_variant_unref(var);
}

int check_power_supply_noti(void)
{
	int block;

	if (vconf_get_bool("db/setting/blockmode_wearable", &block) != 0)
		return 1;
	if (block == 0)
		return 1;
	return 0;
}

static int send_full_noti(enum charge_full_type state)
{
	int ret = 0;
	int noti;
	int retry;

	noti = check_power_supply_noti();

	if (!noti)
		return noti;

	switch (state) {
	case CHARGING_FULL:
		for (retry = RETRY_MAX; retry > 0; retry--) {
			ret = dbus_handle_method_async_with_reply(POPUP_BUS_NAME,
					POPUP_PATH_NOTI,
					POPUP_INTERFACE_NOTI,
					METHOD_FULL_NOTI_ON,
					NULL, NULL, full_noti_cb, -1, NULL);
			if (ret == 0) {
				_D("Created battery full noti.");
				return ret;
			}
		}
		_E("Failed to call dbus method: %d", ret);
	break;
	case CHARGING_NOT_FULL:
		if (noti_id <= 0)
			return -EPERM;
		for (retry = RETRY_MAX; retry > 0; retry--) {
			ret = dbus_handle_method_async_with_reply_var(POPUP_BUS_NAME,
					POPUP_PATH_NOTI,
					POPUP_INTERFACE_NOTI,
					METHOD_FULL_NOTI_OFF,
					g_variant_new("(i)", noti_id),
					noti_off_cb, -1, NULL);
			if (ret == 0) {
				_D("Deleted battery full noti.");
				noti_id = 0;
				return ret;
			}
		}
		_E("Failed to call dbus method: %d", ret);
	break;
	}
	return ret;
}

static void charge_noti_on(GVariant *var, void *user_data, GError *err)
{
	int id = 0;

	if (!var)
		return;

	if (!dh_get_param_from_var(var, "(i)", &id)) {
		_E("Failed to notify charge: no message(%s)", g_variant_get_type_string(var));
		goto out;
	}

	_D("Inserted battery charge noti: %d", id);

out:
	g_variant_unref(var);
}

static int send_charge_noti(void)
{
	int ret = 0;
	int retry;

	for (retry = RETRY_MAX; retry > 0; retry--) {
		ret = dbus_handle_method_async_with_reply(POPUP_BUS_NAME,
				POPUP_PATH_NOTI,
				POPUP_INTERFACE_NOTI,
				METHOD_CHARGE_NOTI_ON,
				NULL, NULL, charge_noti_on, -1, NULL);
		if (ret == 0) {
			_D("Created battery charge noti.");
			return ret;
		}
	}
	_E("Failed to call dbus method: %d", ret);
	return ret;
}

static void power_supply_noti(enum battery_noti_type type, enum battery_noti_status status)
{
	static int charger = CHARGER_DISCHARGING;
	static int full = CHARGING_NOT_FULL;
	int ret;

	if (type == DEVICE_NOTI_BATT_CHARGE) {
		if (status == DEVICE_NOTI_ON && charger == CHARGER_DISCHARGING) {
			send_charge_noti();
			charger = CHARGER_CHARGING;
		} else if (status == DEVICE_NOTI_OFF && charger == CHARGER_CHARGING)
			charger = CHARGER_DISCHARGING;
	} else if (type == DEVICE_NOTI_BATT_FULL) {
		if (status == DEVICE_NOTI_ON && full == CHARGING_NOT_FULL) {
			ret = send_full_noti(CHARGING_FULL);
			if (ret == 0)
				full = CHARGING_FULL;
		} else if (status == DEVICE_NOTI_OFF && full == CHARGING_FULL) {
			ret = send_full_noti(CHARGING_NOT_FULL);
			if (ret == 0)
				full = CHARGING_NOT_FULL;
		}
	}
}

int power_supply_broadcast(char *sig, int status)
{
	static int old;
	static char sig_old[32];
	int ret;
	int chg_on;

	if (!sig) {
		_E("There is no signal name.");
		return -EINVAL;
	}
	if (strncmp(sig_old, sig, strlen(sig)) == 0 && old == status) {
		_D("Skip broadcasting same signal(%s) and status(%d).", sig, status);
		return 0;
	}

	old = status;

	if (battery.charger_charging == CHARGER_DISABLED) {
		ret = sys_get_int(CHARGER_CHARGE_ON_NODE, &chg_on);
		if (ret == 0 && battery.charger_charging != chg_on) {
			ret = sys_set_int(CHARGER_CHARGE_ON_NODE, battery.charger_charging);
			_I("%s to change status with %d", ((ret == 0) ? "success" : "fail"), battery.charger_charging);
		}
		if (strncmp(sig, CHARGE_NOW_SIGNAL, strlen(CHARGE_NOW_SIGNAL)) == 0) {
			_I("Skip signal while charger disabled.");
			return 0;
		}
	}

	snprintf(sig_old, sizeof(sig_old), "%s", sig);

	ret = dbus_handle_emit_dbus_signal(NULL,
						DEVICED_PATH_BATTERY,
						DEVICED_INTERFACE_BATTERY,
						sig,
						g_variant_new("(i)", status));
	if (ret == 0)
		return 1;
	else
		return ret;
}

static void noti_batt_full(void)
{
	static int bat_full_noti;
	int noti;

	if (!battery.charge_full && bat_full_noti == 1) {
		power_supply_noti(DEVICE_NOTI_BATT_FULL, DEVICE_NOTI_OFF);
		bat_full_noti = 0;
		/* off the full charge state */
		device_notify(DEVICE_NOTIFIER_FULLBAT, (void *)&bat_full_noti);
	}
	if (battery.charge_full && bat_full_noti == 0) {
		power_supply_noti(DEVICE_NOTI_BATT_FULL, DEVICE_NOTI_ON);
		bat_full_noti = 1;
		/* turn on LCD, if battery is full charged */
		noti = check_power_supply_noti();
		if (noti) {
			battery_pm_change_internal(INTERNAL_LOCK_BATTERY_FULL, LCD_NORMAL);
		} else
			_I("Block LCD.");

		/* on the full charge state */
		device_notify(DEVICE_NOTIFIER_FULLBAT, (void *)&bat_full_noti);
	}
}

static void check_power_supply(int state)
{
	pm_check_and_change(state);
	if (disp_plgn.update_pm_setting)
		disp_plgn.update_pm_setting(SETTING_CHARGING, state);
}

static void charger_state_send_system_event(int state)
{
	const char *str;

	switch (state) {
	case CHARGE_STATUS_CHARGING:
		str = EVT_VAL_BATTERY_CHARGER_CHARGING;
		break;
	case CHARGE_STATUS_FULL:
	case CHARGE_STATUS_DISCHARGING:
		str = EVT_VAL_BATTERY_CHARGER_DISCHARGING;
		break;
	case CHARGE_STATUS_CONNECTED:
		str = EVT_VAL_BATTERY_CHARGER_CONNECTED;
		break;
	case CHARGE_STATUS_DISCONNECTED:
		str = EVT_VAL_BATTERY_CHARGER_DISCONNECTED;
		break;
	default:
		_E("Invalid parameter: %d", state);
		return;
	}

	_D("System_event: %s", str);

	event_system_send(SYS_EVENT_BATTERY_CHARGER_STATUS, EVT_KEY_BATTERY_CHARGER_STATUS, str);
}

int changed_battery_cf(int status)
{
	if (status == PRESENT_ABNORMAL)
		return launch_system_app(APP_ABNORMAL, 2, APP_KEY_TYPE, DISCONNECT_POPUP);
	return launch_system_app(APP_REMOVE, 2, APP_KEY_TYPE, REMOVE_POPUP);
}

static void update_present(enum battery_noti_status status)
{

	_I("Charge(%d) present(%d)", battery.charge_now, battery.present);

	if (status == DEVICE_NOTI_ON) {
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
		device_notify(DEVICE_NOTIFIER_BATTERY_PRESENT, (void *)&battery.present);
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
	} else {
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
		device_notify(DEVICE_NOTIFIER_BATTERY_PRESENT, (void *)&battery.present);
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, PM_SLEEP_MARGIN);
	}
	changed_battery_cf(battery.present);
}

void remove_health_popup(void)
{
	int ret;

	ret = launch_system_app(APP_REMOVE, 2, APP_KEY_TYPE, REMOVE_POPUP);
	if (ret < 0)
		_E("Failed to launch remove battery popup(%d)", ret);
}

static void update_health(enum battery_noti_status status)
{
	_I("Charge(%d) health(%d)", battery.charge_now, battery.health);

	if (status == DEVICE_NOTI_ON) {
		_I("Popup: Battery health status is not good.");
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
		if (disp_plgn.pm_change_internal)
			disp_plgn.pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
		device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
		if (disp_plgn.pm_lock_internal)
			disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, 0);
		change_health_popup_timer(status);
		if (battery.health == HEALTH_LOW)
			battery_charge_err_low_act(NULL);
		else if (battery.health == HEALTH_HIGH)
			battery_charge_err_high_act(NULL);
	} else {
		battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
		device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
		if (disp_plgn.pm_unlock_internal) {
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, PM_SLEEP_MARGIN);
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
		}
		change_health_popup_timer(status);
		health_status_broadcast();
		abnormal_popup_timer_init();
		remove_health_popup();
	}
}

void update_health_popup_relaunch(void)
{
	int delta;

	_D("update_health_popup_relaunch()");
	clock_gettime(CLOCK_REALTIME, &health_relaunch_end_st);
	delta = health_relaunch_end_st.tv_sec - health_relaunch_start_st.tv_sec;
	if (delta <= 0 || delta >= HEALTH_POPUP_RELAUNCH_TIMEOUT)
		return;

	update_health_popup_relaunch_time((HEALTH_POPUP_RELAUNCH_TIMEOUT - delta));
}

static void update_ovp(enum battery_noti_status status)
{
	static int old = DEVICE_NOTI_OFF;

	if (old == status)
		return;

	old = status;
	_I("Charge(%d) ovp(%d) with lcd(%s)", battery.charge_now, battery.health,
		(status == DEVICE_NOTI_ON) ? "dim" : "normal");

	device_notify(DEVICE_NOTIFIER_BATTERY_OVP, (void *)&battery.health);
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
}

static void check_abnormal_status(void)
{
	if (battery.health == HEALTH_LOW || battery.health == HEALTH_HIGH)
		update_health_popup_relaunch();

	if (old_battery.health != HEALTH_LOW && old_battery.health != HEALTH_HIGH &&
	    (battery.health == HEALTH_LOW || battery.health == HEALTH_HIGH))
		update_health(DEVICE_NOTI_ON);
	else if ((old_battery.health == HEALTH_LOW || old_battery.health == HEALTH_HIGH) &&
		battery.health != HEALTH_LOW && battery.health != HEALTH_HIGH)
		update_health(DEVICE_NOTI_OFF);

	if (old_battery.present != PRESENT_ABNORMAL && battery.present == PRESENT_ABNORMAL)
		update_present(DEVICE_NOTI_ON);
	else if (battery.present != PRESENT_ABNORMAL && old_battery.present == PRESENT_ABNORMAL)
		update_present(DEVICE_NOTI_OFF);

	if (old_battery.health != HEALTH_OVP && battery.health == HEALTH_OVP)
		update_ovp(DEVICE_NOTI_ON);
	else if (battery.health != HEALTH_OVP && old_battery.health == HEALTH_OVP)
		update_ovp(DEVICE_NOTI_OFF);
}

static bool update_online(void)
{
	int chg_on;
	int ret;
	bool broadcast = false;

	if (battery.online <= POWER_SUPPLY_TYPE_BATTERY &&
	    battery.charger_charging == CHARGER_DISABLED) {
		battery.charger_charging = CHARGER_ENABLED;
		ret = sys_get_int(CHARGER_CHARGE_ON_NODE, &chg_on);
		if (ret == 0 && battery.charger_charging != chg_on) {
			ret = sys_set_int(CHARGER_CHARGE_ON_NODE, battery.charger_charging);
			_I("%s to change status with %d", ((ret == 0) ? "success" : "fail"), battery.charger_charging);
		}
	}

	if (battery.online > POWER_SUPPLY_TYPE_BATTERY &&
	    online_status == VCONFKEY_SYSMAN_CHARGER_DISCONNECTED) {
		online_status = VCONFKEY_SYSMAN_CHARGER_CONNECTED;
		extcon_update_count(EXTCON_TA, 1);
		check_power_supply(online_status);
		charger_state_send_system_event(CHARGE_STATUS_CONNECTED);
		if (old_battery.charge_status != battery.charge_status)
			charger_state_send_system_event(battery.charge_status);
		broadcast = true;
	} else if (battery.online <= POWER_SUPPLY_TYPE_BATTERY &&
	    online_status == VCONFKEY_SYSMAN_CHARGER_CONNECTED) {
		online_status = VCONFKEY_SYSMAN_CHARGER_DISCONNECTED;
		check_power_supply(online_status);
		if (old_battery.charge_status != battery.charge_status)
			charger_state_send_system_event(battery.charge_status);
		charger_state_send_system_event(CHARGE_STATUS_DISCONNECTED);
		broadcast = true;
	} else {
		if (old_battery.charge_status != battery.charge_status)
			charger_state_send_system_event(battery.charge_status);
	}

	if (battery.online <= POWER_SUPPLY_TYPE_BATTERY)
		battery.online_type = CHARGER_TYPE_NONE;
	else if (battery.online == CHARGER_WIRELESS_TYPE_BT ||
	    battery.online == CHARGER_WIRELESS_TYPE_3G)
		battery.online_type = CHARGER_TYPE_WIRELESS;
	else if (battery.online == CHARGER_INCOMPATIBLE_TYPE)
		battery.online_type = CHARGER_TYPE_INCOMPATIBLE;
	else
		battery.online_type = CHARGER_TYPE_WIRE;

	return broadcast;
}

static void check_charge_status(const char *env_value)
{
	int len;

	if (env_value == NULL)
		return;

	len = strlen(env_value);
	if (strncmp(env_value, CHARGEFULL_NAME, len) == 0)
		battery.charge_status = CHARGE_STATUS_FULL;
	else if (strncmp(env_value, CHARGENOW_NAME, len) == 0)
		battery.charge_status = CHARGE_STATUS_CHARGING;
	else if (strncmp(env_value, DISCHARGE_NAME, len) == 0)
		battery.charge_status = CHARGE_STATUS_DISCHARGING;
	else if (strncmp(env_value, NOTCHARGE_NAME, len) == 0)
		battery.charge_status = CHARGE_STATUS_NOT_CHARGING;
	else
		battery.charge_status = CHARGE_STATUS_UNKNOWN;

	if (battery.charge_status == CHARGE_STATUS_FULL) {
		battery.charge_full = CHARGING_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_CHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_CHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_DISCHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	} else if (battery.charge_status == CHARGE_STATUS_NOT_CHARGING) {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_ABNORMAL;
	} else {
		battery.charge_full = CHARGING_NOT_FULL;
		battery.charge_now = CHARGER_DISCHARGING;
	}
}

static void check_health_status(const char *env_value)
{
	int len;

	if (env_value == NULL) {
		battery.health = HEALTH_GOOD;
		return;
	}

	snprintf(battery.health_s, sizeof(battery.health_s), "%s", env_value);

	len = strlen(env_value);
	if (strncmp(env_value, OVERHEAT_NAME, len) == 0)
		battery.health = HEALTH_HIGH;
	else if (strncmp(env_value, TEMPCOLD_NAME, len) == 0)
		battery.health = HEALTH_LOW;
	else if (strncmp(env_value, OVERVOLT_NAME, len) == 0)
		battery.health = HEALTH_OVP;
	else if (strncmp(env_value, GOOD_NAME, len) == 0)
		battery.health = HEALTH_GOOD;
	else
		battery.health = HEALTH_NO_OPT;
}

static void check_misc_status(const char *env_value)
{
	if (env_value == NULL) {
		battery.misc = MISC_NONE;
		return;
	}
	battery.misc = atoi(env_value);
}

static void check_online_status(const char *env_value)
{
	if (env_value == NULL)
		return;
	battery.online = atoi(env_value);
}

static void check_present_status(const char *env_value)
{
	if (env_value == NULL) {
		battery.present = PRESENT_NORMAL;
		return;
	}
	battery.present = atoi(env_value);
}

static void check_capacity_status(const char *env_value)
{
	if (env_value == NULL)
		return;
	battery.capacity = atoi(env_value);
}

static void update_capacity_full(void)
{
	static int old;
	int delta;

	if (battery.online <= POWER_SUPPLY_TYPE_BATTERY ||
	    old == 0 ||
	    old >= battery.capacity) {
		old = battery.capacity;
		return;
	}
	delta = battery.capacity - old;
	old = battery.capacity;
	extcon_update_count(EXTCON_BATTERY_FULL, delta);
}

static void update_battery_cycle(void)
{
	static int first = 1;
	static int old;
	int delta;

	if (first) {
		first = 0;
		extcon_update_count(EXTCON_BATTERY_CYCLE, 0);
		return;
	}

	if (battery.online > POWER_SUPPLY_TYPE_BATTERY ||
	    old == 0 ||
	    old <= battery.capacity) {
		old = battery.capacity;
		return;
	}
	delta = old - battery.capacity;
	old = battery.capacity;
	extcon_update_count(EXTCON_BATTERY_CYCLE, delta);
}

static void process_power_supply(void *data)
{
	bool broadcasted = true;
	int lock = -1;
	int ret;

	_D("process_power_supply()");
	if (disp_plgn.pm_lock_internal) {
		lock = disp_plgn.pm_lock_internal(INTERNAL_LOCK_BATTERY, LCD_OFF, STAY_CUR_STATE, 0);
	if (old_battery.charge_now != battery.charge_now || battery.charge_now == CHARGER_ABNORMAL) {
			ret = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CHARGE_NOW, battery.charge_now);
			if (ret < 0)
				_E("Failed to set vconf value for battery charge now: %d", vconf_get_ext_errno());

			if (power_supply_broadcast(CHARGE_NOW_SIGNAL, battery.charge_now) < 0)
				broadcasted = false;
		}
	}

	if (old_battery.charge_full != battery.charge_full)
		if (power_supply_broadcast(CHARGE_FULL_SIGNAL, battery.charge_full) < 0)
			broadcasted = false;

	if (strncmp(old_battery.health_s, battery.health_s, strlen(battery.health_s)))
		if (power_supply_broadcast_str(CHARGE_HEALTH_SIGNAL, battery.health_s) < 0)
			broadcasted = false;

	if (old_battery.capacity != battery.capacity) {
		ret = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CAPACITY, battery.capacity);
		if (ret < 0)
			_E("Failed to set vconf value for battery capacity: %d", vconf_get_ext_errno());

		if (power_supply_broadcast(CHARGE_CAPACITY_SIGNAL, battery.capacity) < 0)
			broadcasted = false;
	}
	battery.charging_level = lowbat_execute(data);
	if (battery.charging_level > 0 && old_battery.charging_level != battery.charging_level) {
		ret = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, battery.charging_level);
		if (ret < 0)
			_E("Failed to set vconf value for battery capacity: %d", vconf_get_ext_errno());

		if (power_supply_broadcast(CHARGE_LEVEL_SIGNAL, battery.charging_level) < 0)
			broadcasted = false;
		}

	if (update_online()) {
		ret = vconf_set_int(VCONFKEY_SYSMAN_CHARGER_STATUS, online_status);
		if (ret < 0)
			_E("Failed to set vconf value for charger status: %d", vconf_get_ext_errno());

		if (power_supply_broadcast(CHARGER_STATUS_SIGNAL, online_status) < 0)
			broadcasted = false;
	}

	if (old_battery.online_type != battery.online_type) {
		if (power_supply_broadcast(CHARGER_TYPE_SIGNAL, battery.online_type) < 0)
		broadcasted = false;
	}

	if (old_battery.misc != battery.misc)
		if (power_supply_broadcast(CHARGE_MISC_EVENT_SIGNAL, battery.misc) < 0)
			broadcasted = false;

	if (old_battery.charge_full != battery.charge_full)
		noti_batt_full();

	if (old_battery.charge_now != battery.charge_now ||
	    old_battery.charge_full != battery.charge_full ||
	    old_battery.capacity != battery.capacity ||
	    old_battery.health != battery.health ||
		old_battery.misc != battery.misc ||
		old_battery.online_type != battery.online_type)
		_I("Signal(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d : %s)",
			broadcasted,
			CHARGER_STATUS_SIGNAL, online_status,
			CHARGER_TYPE_SIGNAL, battery.online_type,
			CHARGE_NOW_SIGNAL, battery.charge_now,
			CHARGE_FULL_SIGNAL, battery.charge_full,
			CHARGE_CAPACITY_SIGNAL, battery.capacity,
			CHARGE_LEVEL_SIGNAL, battery.charging_level,
			CHARGE_MISC_EVENT_SIGNAL, battery.misc,
			CHARGE_HEALTH_SIGNAL, battery.health, battery.health_s);

	old_battery.capacity = battery.capacity;
	old_battery.charging_level = battery.charging_level;
	old_battery.online = battery.online;
	old_battery.online_type = battery.online_type;
	old_battery.charge_status = battery.charge_status;
	old_battery.charge_full = battery.charge_full;

	check_abnormal_status();

	old_battery.health = battery.health;
	old_battery.misc = battery.misc;
	old_battery.present = battery.present;
	snprintf(old_battery.health_s, sizeof(old_battery.health_s), "%s", battery.health_s);

	device_notify(DEVICE_NOTIFIER_POWER_SUPPLY, NULL);
	if (old_battery.charge_now != battery.charge_now) {
		device_notify(DEVICE_NOTIFIER_BATTERY_CHARGING, &battery.charge_now);
		old_battery.charge_now = battery.charge_now;
	}
	update_capacity_full();
	update_battery_cycle();
	if (lock == 0) {
		if (disp_plgn.pm_unlock_internal)
			disp_plgn.pm_unlock_internal(INTERNAL_LOCK_BATTERY, LCD_OFF, PM_SLEEP_MARGIN);
	}
}

static void uevent_power_handler(struct udev_device *dev)
{
	struct udev_list_entry *list_entry;
	const char *env_name;
	const char *env_value;
	bool matched = false;
	int ret;

	udev_list_entry_foreach(list_entry,
			udev_device_get_properties_list_entry(dev)) {
		env_name = udev_list_entry_get_name(list_entry);
		if (!env_name)
			continue;

		if (!strncmp(env_name, CHARGE_NAME, sizeof(CHARGE_NAME))) {
			env_value = udev_list_entry_get_value(list_entry);
			if (!env_value)
				continue;
			if (!strncmp(env_value, BATTERY_NAME,
						sizeof(BATTERY_NAME))) {
				matched = true;
				break;
			}
		}
	}

	if (!matched)
		return;

	env_value = udev_device_get_property_value(dev, CHARGE_STATUS);
	check_charge_status(env_value);
	env_value = udev_device_get_property_value(dev, CHARGE_ONLINE);
	check_online_status(env_value);
	env_value = udev_device_get_property_value(dev, CHARGE_HEALTH);
	check_health_status(env_value);
	env_value = udev_device_get_property_value(dev, CHARGE_MISC_EVENT);
	check_misc_status(env_value);
	env_value = udev_device_get_property_value(dev, CHARGE_PRESENT);
	check_present_status(env_value);
	env_value = udev_device_get_property_value(dev, CAPACITY);
	check_capacity_status(env_value);

	ret = booting_done(NULL);
	if (ret) {
		if (battery.online > POWER_SUPPLY_TYPE_BATTERY)
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_ON);
		else
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_OFF);
	}

	process_power_supply(&battery.capacity);
}

static int battery_state(struct battery_info *info)
{
	static struct battery_status prev_status;

	if (!info) {
		memcpy(&prev_status, &battery, sizeof(battery));
		return 0;
	}

	if (prev_status.capacity == battery.capacity &&
	    prev_status.charge_status == battery.charge_status &&
	    prev_status.charge_full == battery.charge_full &&
	    prev_status.charge_now == battery.charge_now &&
	    prev_status.health == battery.health &&
	    prev_status.present == battery.present &&
	    prev_status.online == battery.online)
		return 0;

	prev_status.capacity = battery.capacity;
	prev_status.charge_status = battery.charge_status;
	prev_status.charge_full = battery.charge_full;
	prev_status.charge_now = battery.charge_now;
	prev_status.health = battery.health;
	prev_status.present = battery.present;
	prev_status.online = battery.online;

	_I("%s(%s) %s(%d) Capa(%d) Hth(%s,%d) Pres(%d) Curr(%d,%d)",
			info->status,
			battery.charge_now == CHARGER_CHARGING ? "Charging"
				: (battery.charge_now == CHARGER_DISCHARGING ? "Discharging" : "Abnormal"),
			info->power_source,
			info->online,
			info->capacity,
			info->health,
			battery.health,
			info->present,
			info->current_now,
			info->current_average);
	return 1;
}

static void battery_changed(struct battery_info *info, void *data)
{
	int ret;

	if (!info) {
		(void)battery_state(NULL);
		return;
	}

	if (info->status) {
		snprintf(battery.status_s, sizeof(battery.status_s),
				"%s", info->status);
		check_charge_status(info->status);
	} else
		battery.status_s[0] = '\0';

	if (info->health) {
		snprintf(battery.health_s, sizeof(battery.health_s),
				"%s", info->health);
		check_health_status(info->health);
	} else
		battery.health_s[0] = '\0';

	if (info->power_source)
		snprintf(battery.power_source_s, sizeof(battery.power_source_s),
				"%s", info->power_source);
	else
		battery.power_source_s[0] = '\0';

	battery.online = info->online;
	battery.present = info->present;
	battery.capacity = info->capacity;
	battery.current_now = info->current_now;
	battery.current_average = info->current_average;
	battery.temperature = info->temperature;

	ret = battery_state(info);
	if (ret != 1)
		return;

	ret = booting_done(NULL);
	if (ret && old_battery.online != battery.online) {
		if (battery.online > POWER_SUPPLY_TYPE_BATTERY && battery.charge_now == CHARGER_CHARGING)
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_ON);
		else
			power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_OFF);
	}
	if (ret == 0) {
		battery.health = HEALTH_GOOD;
		battery.present = PRESENT_NORMAL;
	}

	process_power_supply(&battery.capacity);

}

static void power_supply_status_init(void)
{
	int r;

	if (battery_dev && battery_dev->get_current_state) {
		r = battery_dev->get_current_state(battery_changed, NULL);
		if (r < 0 || battery.capacity < 0) {
			_E("Failed to get battery capacity (capa: %d, ret: %d)", battery.capacity, r);
			return;
		}
	} else {
		r = config_parse(POWER_SUPPLY_UEVENT, load_uevent, &battery);
		if (r < 0) {
			_E("Failed to load %s, %d Use default value.", POWER_SUPPLY_UEVENT, r);
			return;
		}
		battery.health = HEALTH_GOOD;
		battery.present = PRESENT_NORMAL;
		process_power_supply(&battery.capacity);
	}
}

static gboolean power_supply_update(void *data)
{
	power_supply_status_init();
	return G_SOURCE_CONTINUE;
}

static void power_supply_timer_start(void)
{
	_D("Battery init timer during booting.");
	power_timer = g_timeout_add(BATTERY_CHECK_TIMER_INTERVAL,
				power_supply_update, NULL);
	if (power_timer == 0)
		_E("Failed to add battery init timer during booting.");
	else
		_I("Started battery init timer during booting.");
}

static void power_supply_timer_stop(void)
{
	if (power_timer == 0)
		return;
	_I("Stop battery init timer during booting.");
	g_source_remove(power_timer);
	power_timer = 0;
}

static GVariant *dbus_get_charger_status(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, status;

	ret = vconf_get_int(VCONFKEY_SYSMAN_CHARGER_STATUS, &status);
	if (ret < 0) {
		_E("Failed to get vconf value for charger status: %d", vconf_get_ext_errno());
		status = -EIO;
	}
	return g_variant_new("(i)", status);
}

static GVariant *dbus_get_charger_type(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int type = battery.online_type;

	return g_variant_new("(i)", type);
}

static GVariant *dbus_charger_charging(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int val;
	int chg_on;

	g_variant_get(param, "(i)", &val);

	if (val != CHARGER_ENABLED && val != CHARGER_DISABLED) {
		_E("There is no valid input: %d", battery.charger_charging);
		ret = -EINVAL;
		goto out;
	}
	if (battery.online <= POWER_SUPPLY_TYPE_BATTERY) {
		_E("There is no charger(%d) input(%d).", battery.online, battery.charger_charging);
		ret = -EACCES;
		goto out;
	}
	ret = sys_get_int(CHARGER_CHARGE_ON_NODE, &chg_on);
	if (ret == 0 && val != chg_on) {
		battery.charger_charging = val;
		ret = sys_set_int(CHARGER_CHARGE_ON_NODE, battery.charger_charging);
		_I("%s to change status with %d.", ((ret == 0) ? "success" : "fail"), battery.charger_charging);
	} else {
		_I("Skip change status with %d (ret %d prev %d)", val, ret, battery.charger_charging);
	}
out:
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_charge_now(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.charge_now;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_charge_level(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, val;

	ret = vconf_get_int(VCONFKEY_SYSMAN_BATTERY_STATUS_LOW, &val);
	if (ret < 0) {
		_E("Failed to get vconf value for battery status low: %d", vconf_get_ext_errno());
		val = -EIO;
	}

	return g_variant_new("(i)", val);
}

static GVariant *dbus_get_percent(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.capacity;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_percent_raw(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = -ENOTSUP;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_is_full(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.charge_full;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_health(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	const char *str;

	str = battery.health_s;

	return g_variant_new("(s)", str);
}

static GVariant *dbus_get_misc(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = battery.misc;

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_power_supply_handler(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	pid_t pid;
	int ret = 0;
	int argc = 0;
	char *type_str;
	char *argv[6];

	g_variant_get(param, "(sissssss)", &type_str,
					&argc,
					&argv[0],
					&argv[1],
					&argv[2],
					&argv[3],
					&argv[4],
					&argv[5]);

	if (argc < 0) {
		_E("Message is invalid.");
		ret = -EINVAL;
		goto out;
	}

	pid = dbus_connection_get_sender_pid(conn, sender);
	if (kill(pid, 0) == -1) {
		_E("Process(%d) does not exist, dbus ignored.", pid);
		ret = -ESRCH;
		goto out;
	}
	check_capacity_status(argv[0]);
	check_charge_status(argv[1]);
	check_health_status(argv[2]);
	check_online_status(argv[3]);
	check_present_status(argv[4]);
	check_misc_status(argv[5]);
	if (battery_dev)
		battery_changed(NULL, NULL);
	snprintf(battery.status_s, sizeof(battery.status_s), "%s",
			(battery.charge_now == CHARGER_CHARGING) ? "Charging" :
			(battery.charge_now == CHARGER_ABNORMAL) ? "Not Charging" :
			(battery.charge_full == CHARGING_FULL) ? "Full" :
			"Discharging");
	_I("%s(%d) %s(f:%d n:%d) %s(%d) %s(%d) %s(%d) %s(%d)",
		argv[0],
		battery.capacity,
		argv[1],
		battery.charge_full,
		battery.charge_now,
		argv[2],
		battery.health,
		argv[3],
		battery.online,
		argv[4],
		battery.present,
		argv[5],
		battery.misc);

	if (battery.online > POWER_SUPPLY_TYPE_BATTERY)
		power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_ON);
	else
		power_supply_noti(DEVICE_NOTI_BATT_CHARGE, DEVICE_NOTI_OFF);

	process_power_supply(&battery.capacity);
out:
	g_free(type_str);
	g_free(argv[0]);
	g_free(argv[1]);
	g_free(argv[2]);
	g_free(argv[3]);
	g_free(argv[4]);
	g_free(argv[5]);

	return g_variant_new("(i)", ret);
}

static void battery_get_info(struct battery_info *info, void *data)
{
	struct battery_info *bat = data;

	if (!info || !bat)
		return;

	bat->status = strdup(info->status);
	bat->health = strdup(info->health);
	bat->power_source = strdup(info->power_source);
	bat->online = info->online;
	bat->present = info->present;
	bat->capacity = info->capacity;
	bat->current_now = info->current_now;
	bat->current_average = info->current_average;
}

static GVariant *dbus_get_battery_info(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret, val;
	const char *str;
	struct battery_info info = {0, };

	if (battery_dev && battery_dev->get_current_state) {
	ret = battery_dev->get_current_state(battery_get_info, &info);
	if (ret < 0)
		_E("Failed to get battery info: %d", ret);

	battery_changed(&info, NULL);
	free(info.status);
	free(info.health);
	free(info.power_source);
	} else {
		if (vconf_get_int(VCONFKEY_SYSMAN_USB_STATUS, &val) == 0 &&
			val != VCONFKEY_SYSMAN_USB_DISCONNECTED)
			str = POWER_SOURCE_USB;
		else if (vconf_get_int(VCONFKEY_SYSMAN_CHARGER_STATUS, &val) == 0 &&
			val == VCONFKEY_SYSMAN_CHARGER_CONNECTED)
			str = POWER_SOURCE_AC;
		else
			str = POWER_SOURCE_NONE;
		snprintf(battery.power_source_s, sizeof(battery.power_source_s), "%s", str);

		battery.current_now = -1; /* Not supported */
		battery.current_average = -1; /* Not supported */
		ret = 0;
	}

	return g_variant_new("(isssiiiiiiii)", ret,
					battery.status_s,
					battery.health_s,
					battery.power_source_s,
					battery.online,
					battery.present,
					battery.capacity,
					battery.current_now,
					battery.current_average,
					battery.voltage_now,
					battery.voltage_average,
					battery.temperature);
}

static void update_battery_props(void)
{
	if (config_parse(POWER_SUPPLY_UEVENT, load_uevent, &battery) < 0)
		_E("Failed to check power supply uevent.");
}

static GVariant *dbus_get_battery_props(GDBusConnection *conn,
	const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
	GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	update_battery_props();

	return g_variant_new("(iiiiii)",
					battery.capacity,
					battery.temperature,
					battery.voltage_now,
					battery.voltage_average,
					battery.current_now,
					battery.current_average);
}

static const dbus_method_s dbus_methods[] = {
	{ CHARGER_STATUS_SIGNAL,      NULL, "i", dbus_get_charger_status },
	{ CHARGE_NOW_SIGNAL,          NULL, "i", dbus_get_charge_now },
	{ CHARGE_LEVEL_SIGNAL,         NULL, "i", dbus_get_charge_level },
	{ CHARGE_CAPACITY_SIGNAL,     NULL, "i", dbus_get_percent },
	{ CHARGE_CAPACITY_LAW_SIGNAL, NULL, "i", dbus_get_percent_raw },
	{ CHARGE_FULL_SIGNAL,         NULL, "i", dbus_is_full },
	{ CHARGE_HEALTH_SIGNAL,        NULL, "s", dbus_get_health },
	{ POWER_SUBSYSTEM,      "sissssss", "i", dbus_power_supply_handler },
	{ "GetBatteryInfo",           NULL, "isssiiiiiiii", dbus_get_battery_info },
	{ CHARGER_TYPE_SIGNAL,         NULL, "i", dbus_get_charger_type },
	{ "ChargerCharging",            "i", "i", dbus_charger_charging },
	{ CHARGE_BATTERY_PROPERTIES,   NULL, "iiiiii", dbus_get_battery_props },
	{ CHARGE_MISC_EVENT_SIGNAL,    NULL, "i", dbus_get_misc },
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_BATTERY,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static int booting_done(void *data)
{
	static int done;
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_START;

	if (data == NULL)
		return done;
	done = *(int *)data;
	if (done == 0)
		return done;

	power_supply_timer_stop();
	event_handler_state_changed((void *)&state);

	_I("booting done %s(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d) %s(%d : %s)",
			CHARGER_STATUS_SIGNAL, online_status,
			CHARGE_NOW_SIGNAL, battery.charge_now,
			CHARGE_FULL_SIGNAL, battery.charge_full,
			CHARGE_CAPACITY_SIGNAL, battery.capacity,
			CHARGE_LEVEL_SIGNAL, battery.charging_level,
			CHARGE_MISC_EVENT_SIGNAL, battery.misc,
			CHARGE_HEALTH_SIGNAL, battery.health, battery.health_s);

	unregister_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);

	return done;
}

static gboolean update_health_pm_status(void *data)
{
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_OFF);
	if (disp_plgn.pm_unlock_internal) {
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, PM_SLEEP_MARGIN);
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
	}

	return G_SOURCE_REMOVE;
}

static void change_health_popup_timer(enum battery_noti_status status)
{
	if (status == DEVICE_NOTI_ON) {
		health_popup_timer = g_timeout_add_seconds(HEALTH_POPUP_TERMINATE_TIMEOUT, update_health_pm_status, NULL);
		if (health_popup_timer == 0)
			_E("Failed to add battery init timer during booting.");
		else
			_I("Turn off after %d sec by abnormal health condition.", HEALTH_POPUP_TERMINATE_TIMEOUT);
		return;
	}
	health_popup_timer_init(__func__);
}

static void launch_health_popup(void)
{
	_I("Popup - Health is not good(health %s)", (battery.health == HEALTH_HIGH) ? "TempHigh" : "TempLow");
	health_popup_timer_init(__func__);
	device_notify(DEVICE_NOTIFIER_BATTERY_HEALTH, (void *)&battery.health);
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_DIM);
	if (disp_plgn.pm_unlock_internal)
		disp_plgn.pm_unlock_internal(INTERNAL_LOCK_POPUP, LCD_OFF, PM_SLEEP_MARGIN);
	if (disp_plgn.pm_lock_internal)
		disp_plgn.pm_lock_internal(INTERNAL_LOCK_POPUP, LCD_DIM, STAY_CUR_STATE, LCD_DIM_TIME_IN_BATTERY_HEALTH);
	change_health_popup_timer(DEVICE_NOTI_ON);
	if (battery.health == HEALTH_LOW)
		battery_charge_err_low_act(NULL);
	else if (battery.health == HEALTH_HIGH)
		battery_charge_err_high_act(NULL);
}

bool battery_do_not_disturb(void)
{
	int block = 0;
	int r;

	if (display_changed(NULL) == S_LCDOFF) {
		r = vconf_get_bool(VCONFKEY_SETAPPL_BLOCKMODE_WEARABLE_BOOL, &block);
		if (r < 0)
			_E("Failed to set vconf value for blockmode wearable: %d", vconf_get_ext_errno());
		r = vconf_get_bool(VCONFKEY_SETAPPL_THEATER_MODE_ENABLE, &block);
		if (r < 0)
			_E("Failed to set vconf value for theator mode enable: %d", vconf_get_ext_errno());
		r = vconf_get_bool(VCONFKEY_SETAPPL_GOODNIGHT_MODE_ENABLE, &block);
		if (r < 0)
			_E("Failed to set vconf value for goodnight mode enable: %d", vconf_get_ext_errno());
	}

	if (block != 0) {
		_I("Skip lcd and popup(block %d).", block);
		return true;
	}

	return false;
}

static int battery_pm_change_internal(int pid, int s_bits)
{
	if (battery_do_not_disturb())
		return 0;

	if (disp_plgn.pm_change_internal)
		disp_plgn.pm_change_internal(pid, s_bits);

	return 0;
}

static gboolean change_health_pm_normal(void *data)
{
	battery_pm_change_internal(INTERNAL_LOCK_POPUP, LCD_NORMAL);
	return G_SOURCE_REMOVE;
}

static void update_health_popup_relaunch_time(int timeout)
{
	if (timeout <= 0)
		return;

	health_popup_timer_init(__func__);
	health_popup_timer = g_timeout_add(timeout, change_health_pm_normal, NULL);
	if (health_popup_timer == 0)
		_E("Failed to add battery init timer during booting.");
	else
		_D("Relaunch timeout with %d sec.", timeout);
}

static void health_popup_timer_init(const char *func)
{
	if (health_popup_timer == 0)
		return;

	g_source_remove(health_popup_timer);
	health_popup_timer = 0;
	_D("init by %s", func);
}

static int display_changed(void *data)
{
	static enum state_t old;
	static enum state_t state;

	if (!data)
		return state;

	state = *(int *)data;

	if (battery.health != HEALTH_LOW && battery.health != HEALTH_HIGH)
		goto out;

	if (old != S_LCDOFF || state != S_LCDON) {
		if (state == S_LCDOFF) {
			health_popup_timer_init(__func__);
			clock_gettime(CLOCK_REALTIME, &health_relaunch_start_st);
			update_health_popup_relaunch_time(HEALTH_POPUP_RELAUNCH_TIMEOUT);
		}
		goto out;
	}
	launch_health_popup();

out:
	old = state;
	return 0;
}

static int load_uevent(struct parse_result *result, void *user_data)
{
	struct battery_status *info = user_data;

	if (!info)
		return -EINVAL;

	if (MATCH(result->name, CHARGE_STATUS)) {
		if (strstr(result->value, "Charging")) {
			info->charge_now = CHARGER_CHARGING;
			info->charge_full = CHARGING_NOT_FULL;
		} else if (strstr(result->value, "Discharging")) {
			info->charge_now = CHARGER_DISCHARGING;
			info->charge_full = CHARGING_NOT_FULL;
		} else if (strstr(result->value, "Full")) {
			info->charge_now = CHARGER_DISCHARGING;
			info->charge_full = CHARGING_FULL;
		} else if (strstr(result->value, "Not charging")) {
			info->charge_now = CHARGER_ABNORMAL;
			info->charge_full = CHARGING_NOT_FULL;
		}
		snprintf(info->status_s, sizeof(info->status_s), "%s", result->value);
	} else if (MATCH(result->name, CAPACITY)) {
		info->capacity = atoi(result->value);
	} else if (MATCH(result->name, TEMPERATURE)) {
		info->temperature = atoi(result->value);
	} else if (MATCH(result->name, VOLTAGE_NOW)) {
		info->voltage_now = atoi(result->value);
	} else if (MATCH(result->name, VOLTAGE_AVG)) {
		info->voltage_average = atoi(result->value);
	} else if (MATCH(result->name, CURRENT_NOW)) {
		info->current_now = atoi(result->value);
	} else if (MATCH(result->name, CURRENT_AVG)) {
		info->current_average = atoi(result->value);
	} else if (MATCH(result->name, CHARGE_HEALTH)) {
		snprintf(info->health_s, sizeof(info->health_s), "%s", result->value);
	}

	return 0;
}

static int power_supply_probe(void *data)
{
	struct hw_info *info;
	int ret, r;

	if (battery_dev)
		return 0;

	ret = hw_get_info(BATTERY_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) { /* There is no HAL for battery */
		if (access(POWER_PATH, R_OK) == 0)
			return 0; /* Just power_supply uevent is used */
		goto out;
	}

	if (!info->open) {
		_E("Failed to open battery device: open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, NULL, (struct hw_common**)&battery_dev);
	if (ret < 0) {
		_E("Failed to get battery device structure: %d", ret);
		return ret;
	}

	if (!battery_dev || !battery_dev->get_current_state) {
		_E("get_current_state() is not supported by the Battery HAL.");
		return -ENODEV;
	}

	_I("Battery device structure load success.");
	return 0;

out:
	/**
	 * Set battery vconf as -ENOTSUP
	 * These vconf key used by runtime-info and capi-system-device.
	 */
	r = vconf_set_int(VCONFKEY_SYSMAN_CHARGER_STATUS, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for charger status: %d", vconf_get_ext_errno());
	r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_CHARGE_NOW, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for battery charge now: %d", vconf_get_ext_errno());
	r = vconf_set_int(VCONFKEY_SYSMAN_BATTERY_LEVEL_STATUS, -ENOTSUP);
	if (r < 0)
		_E("Failed to set vconf value for battery level status: %d", vconf_get_ext_errno());
	_I("There is no battery device: %d", ret);
	return -ENODEV;
}

static void add_power_supply_handler(void)
{
	int ret;

	if (battery_dev) {
		if (battery_dev->register_changed_event)
			battery_dev->register_changed_event(battery_changed, NULL);
		if (battery_dev->get_current_state)
			battery_dev->get_current_state(battery_changed, NULL);
	} else {
		ret = config_parse(POWER_SUPPLY_UEVENT, load_uevent, &battery);
		if (ret < 0)
			_E("Failed to load %s, %d Use default value!", POWER_SUPPLY_UEVENT, ret);

		/* register power subsystem */
		register_kernel_uevent_control(&uh);
	}
}

static void remove_power_supply_handler(void)
{
	if (battery_dev)
		battery_dev->unregister_changed_event(battery_changed);
	else
		unregister_kernel_uevent_control(&uh);
}

static int event_handler_state_changed(void *data)
{
	static device_notifier_state_e old = DEVICE_NOTIFIER_STATE_STOP;
	device_notifier_state_e state = *(device_notifier_state_e *)data;

	if (old == state)
		return 0;

	old = state;

	if (state == DEVICE_NOTIFIER_STATE_START)
		add_power_supply_handler();
	else if (state == DEVICE_NOTIFIER_STATE_STOP)
		remove_power_supply_handler();

	return 0;
}

static void power_supply_init(void *data)
{
	int ret;

	memset(&battery, 0, sizeof(struct battery_status));
	memset(&old_battery, 0, sizeof(struct battery_status));
	battery.capacity = -1;
	battery.charger_charging = CHARGER_ENABLED;
	battery.misc = MISC_NONE;

	/* process check battery timer until booting done */
	power_supply_timer_start();

	register_notifier(DEVICE_NOTIFIER_BOOTING_DONE, booting_done);
	register_notifier(DEVICE_NOTIFIER_LCD, display_changed);
	register_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_BATTERY, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);


	ret = subscribe_dbus_signal(NULL, DEVICED_PATH_SYSNOTI,
		DEVICED_INTERFACE_SYSNOTI, SIGNAL_CHARGEERR_RESPONSE, abnormal_popup_dbus_signal_handler, NULL, NULL);
	if (ret <= 0)
		_E("Failed to init dbus signal: %d", ret);
}

static void power_supply_exit(void *data)
{
	device_notifier_state_e state = DEVICE_NOTIFIER_STATE_STOP;

	unregister_notifier(DEVICE_NOTIFIER_EVENT_HANDLER, event_handler_state_changed);

	event_handler_state_changed((void *)&state);
}

static const struct device_ops power_supply_ops = {
	DECLARE_NAME_LEN("power_supply"),
	.probe    = power_supply_probe,
	.init     = power_supply_init,
	.exit     = power_supply_exit,
};

DEVICE_OPS_REGISTER(&power_supply_ops)
