 /*
  * deviced
  *
  * Copyright (c) 2019 Samsung Electronics Co., Ltd. All rights reserved.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  * http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <glib.h>
#include <wayland-client.h>
#include <tizen-extension-client-protocol.h>
#include <tizen-dpms-client-protocol.h>

#include "core/log.h"
#include "display/util.h"
#include "display/display-dpms.h"

#define DPMS_GET_TIMEOUT	5  /* second */

typedef struct {
	GSource gsource;
	GPollFD gfd;
	gboolean prepared;

	struct wl_display *wl_disp;
	struct wl_registry *wl_registry;
	struct wl_output *wl_output;
	struct tizen_dpms_manager *tz_dpms_mng;

	uint32_t dpms_mode;
	bool connected;
} wl_glib_info;

wl_glib_info *set_client = NULL;
wl_glib_info *get_client = NULL;

static bool wm_is_ready;

static GMainLoop *dpms_loop;
static GMainContext *dpms_context;
static bool dpms_get_state_done;

static void dpms_handle_set_state(void *data, struct tizen_dpms_manager *tz_dpms, uint32_t mode, uint32_t error)
{
	_D("Dpms_state_cb set - mode:%u, error:%u", mode, error);
	set_client->dpms_mode = mode;
}

static void dpms_handle_get_state(void *data, struct tizen_dpms_manager *tz_dpms, uint32_t mode, uint32_t error)
{
	_D("Dpms_state_cb get - mode:%u, error:%u", mode, error);
	get_client->dpms_mode = mode;

	dpms_get_state_done = true;
	if (g_main_loop_is_running(dpms_loop))
		g_main_loop_quit(dpms_loop);
}

static gboolean dpms_loop_quit(void *data)
{
	if (g_main_loop_is_running(dpms_loop))
		g_main_loop_quit(dpms_loop);

	return G_SOURCE_REMOVE;
}

static void disconnect_interface_with_dpms(wl_glib_info *info)
{
	if (!info) {
		_E("Info is NULL.");
		return;
	}

	if (!info->connected)
		return;

	info->connected = false;

	if (info->tz_dpms_mng) {
		tizen_dpms_manager_destroy(info->tz_dpms_mng);
		info->tz_dpms_mng = NULL;
	}

	if (info->wl_output) {
		wl_output_destroy(info->wl_output);
		info->wl_output = NULL;
	}

	if (info->wl_registry) {
		wl_registry_destroy(info->wl_registry);
		info->wl_registry= NULL;
	}

	if (info->wl_disp) {
		wl_display_disconnect(info->wl_disp);
		info->wl_disp= NULL;
	}

	if (info == set_client) {
		set_client = NULL;
		_D("Disconnect set client.");
	} else if (info == get_client) {
		get_client = NULL;
		_D("Disconnect get clinet.");
	}
}

static const struct tizen_dpms_manager_listener dpms_listener = {
	dpms_handle_set_state,
	dpms_handle_get_state
};

static void handle_global(void *data, struct wl_registry *registry,
				uint32_t name, const char *interface, uint32_t version)
{
	wl_glib_info *info;
	if (!data) {
		_E("Data is NULL.");
		return;
	}
	info = (wl_glib_info *)data;

	if (strcmp(interface, "wl_output") == 0) {
		info->wl_output= wl_registry_bind(registry, name, &wl_output_interface, 2);
		if (!info->wl_output) {
			_E("Failed to bind registry.");
			exit_dpms();
			return;
		}
		_D("Get wl_output.");
	} else if (strcmp(interface, "tizen_dpms_manager") == 0) {
		info->tz_dpms_mng = wl_registry_bind(registry, name, &tizen_dpms_manager_interface, 1);
		if (!info->tz_dpms_mng) {
			_E("Failed to bind registry.");
			exit_dpms();
			return;
		}

		tizen_dpms_manager_add_listener(info->tz_dpms_mng, &dpms_listener, info);
		_D("Get dpms manager.");
	}
}

static void handle_global_remove(void *data, struct wl_registry *registry, uint32_t name)
{

}

static const struct wl_registry_listener registry_listener = {
	handle_global,
	handle_global_remove
};

static int
wl_glib_wayland_init(wl_glib_info *info)
{
	int ret;

	info->wl_disp = wl_display_connect(NULL);
	if (info->wl_disp == NULL) {
		_E("Failed to connect display.");
		goto err;
	}

	info->wl_registry = wl_display_get_registry(info->wl_disp);
	if (info->wl_registry == NULL) {
		_E("Failed to get display registry.");
		goto err;
	}

	ret = wl_registry_add_listener(info->wl_registry, &registry_listener, info);
	if (ret) {
		_E("Failed to add registry listener.");
		goto err;
	}

	ret = wl_display_roundtrip(info->wl_disp);
	if (ret < 0) {
		_E("Failed to roundtrip display.");
		goto err;
	}

	if (!info->wl_output) {
		_E("Failed to get wl_output.");
		goto err;
	}

	if (!info->tz_dpms_mng) {
		_E("Failed to get dpms protocol.");
		goto err;
	}

	return 0;

err:
	return -1;
}

static gboolean
wl_glib_client_prepare(GSource *source, gint *time)
{
	wl_glib_info *info;

	if (!source) {
		_E("Source is NULL.");
		return FALSE;
	}
	info = (wl_glib_info *)source;

	/* If info is already prepared, do nothing in this function. */
	if (info->prepared == TRUE)
		return FALSE;

	while (wl_display_prepare_read(info->wl_disp) != 0) {
		if (wl_display_dispatch_pending(info->wl_disp) == -1)
			_E("Failed to dispatch display pending.");
	}

	info->prepared = TRUE;

	wl_display_flush(info->wl_disp);
	*time = -1;

	return FALSE;
}

static gboolean
wl_glib_client_check(GSource *source)
{
	wl_glib_info *info;
	gboolean ret = FALSE;
	int wl_ret = 0;

	if (!source) {
		_E("Source is NULL.");
		return FALSE;
	}
	info = (wl_glib_info *)source;
	if (g_source_is_destroyed(source)) {
		if (info && info->wl_disp && info->prepared)
			wl_display_cancel_read(info->wl_disp);
		_E("Display source(%p) already destroyed.", source);
		return ret;
	}

	if (!info->prepared)
		return ret;

	if (info->gfd.revents & (G_IO_HUP|G_IO_ERR)) {
		wl_display_cancel_read(info->wl_disp);
		goto err;
	}

	if (info->gfd.revents & G_IO_IN) {
		wl_ret = wl_display_read_events(info->wl_disp);
		if (wl_ret < 0) {
			_E("Failed to read display events: %d", wl_ret);
			goto err;
		}
		ret = TRUE;
	} else
		wl_display_cancel_read(info->wl_disp);

	info->prepared = FALSE;

	return ret;

err:
	_E("Get ERROR in check: revent:0x%x, wl_err:%d", info->gfd.revents, wl_ret);

	info->prepared = FALSE;
	g_source_destroy(source);

	return FALSE;
}

static gboolean
wl_glib_client_dispatch(GSource *source, GSourceFunc cb, gpointer data)
{
	wl_glib_info *info;
	int ret = 0;

	if (!source) {
		_E("Source is NULL.");
		return G_SOURCE_REMOVE;
	}
	info = (wl_glib_info *)source;

	if (g_source_is_destroyed(source)) {
		_E("Display source(%p) already destroyed.", source);
		return G_SOURCE_REMOVE;
	}

	if (info->gfd.revents & (G_IO_HUP|G_IO_ERR))
		goto err;

	if (info->gfd.revents & G_IO_IN) {
		ret = wl_display_dispatch_pending(info->wl_disp);
		if (ret == -1) {
			_E("Failed to dispatch display pending: %d", ret);
			goto err;
		}
	}

	ret = wl_display_flush(info->wl_disp);
	if (ret == -1) {
		_E("Failed to flush display: %d", ret);
		goto err;
	}

	return G_SOURCE_CONTINUE;

err:
	_E("Err in dispatch: revent:0x%x, wl_err:%d", info->gfd.revents, ret);

	return G_SOURCE_REMOVE;
}

static void
wl_glib_client_finalize(GSource *source)
{
	if (!source) {
		_E("Source is NULL.");
		return;
	}

	disconnect_interface_with_dpms((wl_glib_info *)source);
	return;
}

static GSourceFuncs wl_glib_client_funcs = {
	.prepare = wl_glib_client_prepare,
	.check = wl_glib_client_check,
	.dispatch = wl_glib_client_dispatch,
	.finalize = wl_glib_client_finalize,
};

static void wl_glib_client_init(wl_glib_info **client, GMainContext *context)
{
	int ret;
	wl_glib_info *info;

	info = (wl_glib_info *)g_source_new(&wl_glib_client_funcs, sizeof(wl_glib_info));
	if (info == NULL) {
		_E("Failed to make g_source_new.");
		goto err;
	}

	ret = wl_glib_wayland_init(info);
	if (ret < 0) {
		_E("Failed to init wayland.");
		goto err;
	}

	info->gfd.fd = wl_display_get_fd(info->wl_disp);
	if (info->gfd.fd < 0) {
		_E("Failed to get display fd");
		goto err;
	}

	info->gfd.events = G_IO_IN | G_IO_ERR;
	info->gfd.revents = 0;

	g_source_add_poll(&info->gsource, &info->gfd);
	g_source_attach(&info->gsource, context);
	g_source_unref(&info->gsource);

	info->connected = true;
	*client = info;
	return;

err:
	if (info)
		g_source_unref(&info->gsource);

	*client = NULL;
}

static bool check_wm_ready(void)
{
	if (access("/run/.wm_ready", F_OK) == 0) {
		_I("Window manager is ready.");
		wm_is_ready = true;
		return true;
	}

	_I("Window manager is not ready.");
	return false;
}

static bool dpms_is_available(void)
{
	/* Check window manager */
	if (!wm_is_ready) {
		_E("Window manager is not ready");
		return false;
	}

	return set_client && get_client;
}

void dpms_set_state(int on)
{
	uint32_t mode;

	if (!dpms_is_available()) {
		_E("Dpms is not available.");
		return;
	}

	mode = (uint32_t)on;
	tizen_dpms_manager_set_dpms(set_client->tz_dpms_mng, set_client->wl_output, mode);
	wl_display_flush(set_client->wl_disp);
}

int dpms_get_state(void)
{
	GSource* timeout;

	if (!dpms_is_available()) {
		_E("Dpms is not available.");
		return -EINVAL;
	}

	// clear pending events
	while (g_main_context_pending(dpms_context))
		g_main_context_iteration(dpms_context, true);

	dpms_get_state_done = false;

	// request for a dpms state to dpms server
	tizen_dpms_manager_get_dpms(get_client->tz_dpms_mng, get_client->wl_output);
	wl_display_flush(get_client->wl_disp);

	// Set a response timeout
	timeout = g_timeout_source_new_seconds(DPMS_GET_TIMEOUT);
	g_source_set_callback(timeout, dpms_loop_quit, NULL, NULL);
	g_source_attach(timeout, dpms_context);

	// Start dpms_loop, waiting for a event
	// dpms_loop stops running when
	// 1. dpms_handle_get_state is invoked (by dpms server as a dpms response)
	// 2. dpms_loop_quit is invoked (by timeout event)
	g_main_loop_run(dpms_loop);
	g_source_destroy(timeout);
	g_source_unref(timeout);

	if (dpms_get_state_done) {
		return (int)get_client->dpms_mode;
	} else {
		_E("No DPMS response for %d seconds.", DPMS_GET_TIMEOUT);
		return -ETIMEDOUT;
	}
}

/*
 * Make two dpms client,
 * each sends only set or get requests.
 *
 * set_client is attached to the default context,
 * whereas get_client is attached to the dpms_context.
 *
 * return true only if both connection established successfully.
 */
bool init_dpms(void)
{
	if (set_client && get_client)
		return true;

	if (!check_wm_ready())
		return false;

	_I("Init dpms.");
	/* client for set request */
	if (!set_client) {
		wl_glib_client_init(&set_client, NULL);

		if (!set_client) {
			_E("Failed to init set client.");
			exit_dpms();
			return false;
		}
		_I("Successfully initialized dpms set client.");
	}

	/* client for get request */
	if (!get_client) {
		dpms_context = g_main_context_new();
		dpms_loop = g_main_loop_new(dpms_context, false);

		wl_glib_client_init(&get_client, dpms_context);

		if (!get_client) {
			_E("Failed to init get clinet.");
			exit_dpms();
			return false;
		}
		_I("Successfully initialized dpms get client.");
	}

	return true;
}

void exit_dpms(void)
{
	_I("Exit dpms.");
	if (dpms_loop) {
		g_main_loop_unref(dpms_loop);
		dpms_loop = NULL;
	}

	if (dpms_context) {
		g_main_context_unref(dpms_context);
		dpms_context = NULL;
	}

	if (set_client && !g_source_is_destroyed((GSource *)set_client))
		g_source_destroy((GSource *)set_client);

	if (get_client && !g_source_is_destroyed((GSource *)get_client))
		g_source_destroy((GSource *)get_client);
}
