/*
 * deviced
 *
 * Copyright (c) 2012 - 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */


#include <stdio.h>
#include <errno.h>
#include <assert.h>
#include <glib.h>
#include <device-node.h>

#include <vconf.h>
#include <hw/led.h>
#include <libsyscommon/dbus-system.h>

#include "core/log.h"
#include "core/list.h"
#include "core/common.h"
#include "core/devices.h"
#include "core/device-notifier.h"

struct rgb_request {
	pid_t pid;
	struct led_state state;
};

/* enum values defined in capi-system-device */
typedef enum {
	LED_CUSTOM_DUTY_ON = 1 << 0,                /**< blink LED */
	LED_CUSTOM_DEFAULT = (LED_CUSTOM_DUTY_ON),  /**< Default flag */
} led_custom_flags;

static struct led_device *rgb_dev;
static dd_list *rgb_list;
static struct rgb_request *top;
static guint timer;

static int find_rgb_req(pid_t pid, struct rgb_request **request)
{
	dd_list *l;
	struct rgb_request *req;

	DD_LIST_FOREACH(rgb_list, l, req) {
		if (req->pid == pid) {
			*request = req;
			return 0;
		}
	}
	return -ENOENT;
}

static int remove_req_from_list(pid_t pid)
{
	dd_list *l;
	struct rgb_request *req;

	DD_LIST_FOREACH(rgb_list, l, req) {
		if (req->pid == pid) {
			_I("Remove rgb play request from pid(%d).", pid);
			if (req == top)
				top = NULL;
			DD_LIST_REMOVE(rgb_list, req);
			free(req);
			return 0;
		}
	}

	return -EINVAL;
}

static int add_req_to_list(pid_t pid, struct led_state *state)
{
	int ret;
	struct rgb_request *req;

	if (!state)
		return -EINVAL;

	ret = find_rgb_req(pid, &req);
	if (ret < 0) {
		req = calloc(1, sizeof(struct rgb_request));
		if (!req) {
			_E("Failed to call calloc().");
			return -ENOMEM;
		}
		req->pid = pid;
		DD_LIST_APPEND(rgb_list, req);
	}

	req->state.color = state->color;
	req->state.duty_on = state->duty_on;
	req->state.duty_off = state->duty_off;
	req->state.type = state->type;

	top = req;

	return 0;
}

static int rgb_play(struct led_state *state)
{
	struct led_state off = { 0, };

	if (!rgb_dev) {
		_E("There is NO HAL.");
		return -ENODEV;
	}

	if (!rgb_dev->set_state) {
		_E("LED HAL does not support set_state().");
		return -ENOTSUP;
	}

	if (state)
		return rgb_dev->set_state(state);

	return rgb_dev->set_state(&off);
}

static gboolean remove_not_handled_req(gpointer data)
{
	dd_list *l;
	int ret = 0;
	struct rgb_request *req;

	DD_LIST_FOREACH(rgb_list, l, req) {
		if (kill(req->pid, 0) == -1) {
			ret = remove_req_from_list(req->pid);
			if (ret < 0)
				_E("Failed to remove rgb request: %d", ret);
		}
}

	if (!top) {
		ret = rgb_play(NULL);
		if (ret < 0)
			_E("Failed to turn off removed request: %d", ret);
	}

	if (DD_LIST_LENGTH(rgb_list) == 0) {
		timer = 0;
		return G_SOURCE_REMOVE;
	}

	return G_SOURCE_CONTINUE;
}

static void add_timer_to_check_invalid_req(void)
{
	if (timer > 0)
		return;

	timer = g_timeout_add(5000, remove_not_handled_req, NULL);
	if (timer == 0)
		_E("Failed to add timer.");
}

static int get_led_num(void)
{
	static bool initialized = false;
	static int num_of_leds = -1;
	int ret;

	if (initialized)
		return num_of_leds;

	if (!rgb_dev) {
		_E("There is NO HAL.");
		ret = -ENODEV;
		goto out;
	}

	if (!rgb_dev->get_number) {
		_E("LED HAL does not support get_number().");
		ret = -ENOTSUP;
		goto out;
	}

	ret = rgb_dev->get_number();
	if (ret < 0) {
		_E("Failed to get LED number: %d", ret);
		goto out;
	}

	num_of_leds = ret;
	initialized = true;

out:
	return ret;
}

static GVariant *dbus_playcustom(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	int on, off;
	unsigned int color, flags;
	struct led_state state;
	pid_t pid;

	g_variant_get(param, "(iiuu)", &on, &off, &color, &flags);

	pid = dbus_connection_get_sender_pid(conn, sender);

	_I("Pid(%d) led play custom on=%d off=%d color=%x.", pid, on, off, color);

	state.color = color;
	state.duty_on = on;
	state.duty_off = off;
	if (flags & LED_CUSTOM_DUTY_ON)
		state.type = LED_TYPE_BLINK;
	else
		state.type = LED_TYPE_MANUAL;

	ret = rgb_play(&state);
	if (ret < 0) {
		_E("Failed to play LED notification: %d", ret);
		goto out;
	}

	add_req_to_list(pid, &state);
	add_timer_to_check_invalid_req();

out:
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_stopcustom(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;
	pid_t pid;

	pid = dbus_connection_get_sender_pid(conn, sender);

	_I("Pid(%d) led stop custom.", pid);

	ret = remove_req_from_list(pid);
	if (ret < 0) {
		_E("There is no request from pid(%d) to remove.", pid);
		goto out;
	}

	if (!top) {
		ret = rgb_play(NULL);
		if (ret < 0) {
			_E("Failed to play LED notification: %d", ret);
			goto out;
		}
	}

out:
	return g_variant_new("(i)", ret);
}

static GVariant *dbus_get_num_of_leds(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret;

	ret = get_led_num();

	return g_variant_new("(i)", ret);
}

static GVariant *dbus_multi_led_control(GDBusConnection *conn,
		const gchar *sender, const gchar *path, const gchar *iface, const gchar *name,
		GVariant *param, GDBusMethodInvocation *invocation, gpointer user_data)
{
	int ret = 0;
	int num_of_leds, size, i;
	unsigned int *color = NULL;
	GVariantIter *iter = NULL;
	pid_t pid;
	struct led_state state;

	pid = dbus_connection_get_sender_pid(conn, sender);

	_I("Pid(%d) multi led play control.", pid);

	ret = get_led_num();
	if (ret < 0)
		goto out;
	else
		num_of_leds = ret;

	g_variant_get(param, "(au)", &iter);
	size = g_variant_iter_n_children(iter);
	if (num_of_leds != size) {
		_E("Led number is not matched.");
		ret = -EINVAL;
		goto out;
	}

	color = (unsigned int*)malloc(sizeof(unsigned int) * size);
	if (!color) {
		ret = -ENOMEM;
		goto fail;
	}

	i = 0;
	state.duty_on = 0;
	state.duty_off = 0;
	state.type = LED_TYPE_MANUAL;

	while (g_variant_iter_next(iter, "u", &color[i])) {
		state.color = color[i];
		ret = rgb_play(&state);
		if (ret < 0)
			_E("Failed to play LED notification: %d.", ret);
		++i;
	}

out:
	if (color)
		free(color);
fail:
	if (iter)
		g_variant_iter_free(iter);

	return g_variant_new("(i)", ret);
}

static const dbus_method_s dbus_methods[] = {
	{ "playcustom",        "iiuu",  "i",  dbus_playcustom },
	{ "stopcustom",        NULL,    "i",  dbus_stopcustom },
	{ "GetNumOfLeds",      NULL,    "i",  dbus_get_num_of_leds},
	{ "multi_led_control",  "au",    "i",  dbus_multi_led_control },
	/* Add methods here */
};

static const dbus_interface_u dbus_interface = {
	.oh = NULL,
	.name = DEVICED_INTERFACE_LED,
	.methods = dbus_methods,
	.nr_methods = ARRAY_SIZE(dbus_methods),
};

static void rgb_init(void *data)
{
	int ret;

	ret = dbus_handle_add_dbus_object(NULL, DEVICED_PATH_LED, &dbus_interface);
	if (ret < 0)
		_E("Failed to init dbus method: %d", ret);

	ret = rgb_play(NULL);
	if (ret < 0)
		_E("Failed to turn off rgb play during init.");
}

static void rgb_exit(void *data)
{
	int ret = rgb_play(NULL);
	if (ret < 0)
		_E("Failed to turn off rgb play during exit.");
}

static int rgb_probe(void *data)
{
	struct hw_info *info;
	int ret;

	if (rgb_dev)
		return 0;

	ret = hw_get_info(LED_HARDWARE_DEVICE_ID,
			(const struct hw_info **)&info);
	if (ret < 0) {
		_E("There is no LED HAL.");
		return -ENOTSUP;
	}

	if (!info->open) {
		_E("Failed to open led device: open(NULL)");
		return -ENODEV;
	}

	ret = info->open(info, LED_ID_NOTIFICATION, (struct hw_common **)&rgb_dev);
	if (ret < 0) {
		_E("Failed to get LED NOTIFICATION device structure: %d", ret);
		return ret;
	}

	_I("LED NOTIFICATION device structure load success.");
	return 0;
}

static const struct device_ops rgbled_device_ops = {
	DECLARE_NAME_LEN("rgbled"),
	.probe    = rgb_probe,
	.init     = rgb_init,
	.exit     = rgb_exit,
};

DEVICE_OPS_REGISTER(&rgbled_device_ops)
